//
//  ATMPPR_Security.m
//  testFypController
//
//  Created by tyler on 3/19/10.
//  Copyright 2010 Atomic Pepper LLC. All rights reserved.
//
//  Triprism Inc., is hereby granted a worldwide, perpetual, fully
//  paid license to use, modify, copy, compile and sublicense the
//  contents of this file.
//

#import "ATMPPR_Security.h"
#include <sys/mman.h>


@interface ATMPPR_SecureSettings ()

@property (nonatomic, retain)     NSMutableDictionary * privateDataStore;
@property (nonatomic, copy)       NSString * privateFilePath;
@property (nonatomic, retain)     NSError * privateLastError;
@property (nonatomic, assign)     BOOL needsSaving;

@end


@implementation ATMPPR_SecureSettings

@synthesize privateDataStore = dataStore_;
@synthesize privateFilePath = filePath_;
@synthesize needsSaving = needsSaving_;
@synthesize privateLastError = lastError_;


id  g_mutablePersistentDict = nil;
id  g_sharedInstance = nil;
BOOL  g_didRestoreFromPrevious = NO;
NSError * g_lastError = nil;

#pragma mark --- start of public Functions ---
static void showErr(int code)
{
    fprintf(stderr,"STATE: +error\n" "ERROR: code %d. Reinstall the software.\n", code);
}
static void fatal(int code)
{
    showErr(code);
	//    sleep(3600);
	//    exit(1);
}

#pragma mark --- static session data and private functions ----

static CSSM_VERSION vers = {2, 0};
static const CSSM_GUID testGuid = { 0xFADE, 0, 0, { 1,2,3,4,5,6,7,0 }};

/*
 * Standard app-level memory functions required by CDSA.
 */
void * appMalloc (CSSM_SIZE size, void *allocRef) {
	return( malloc(size) );
}
void appFree (void *mem_ptr, void *allocRef) {
	free(mem_ptr);
 	return;
}
void * appRealloc (void *ptr, CSSM_SIZE size, void *allocRef) {
	return( realloc( ptr, size ) );
}
void * appCalloc (uint32 num, CSSM_SIZE size, void *allocRef) {
	return( calloc( num, size ) );
}
static CSSM_API_MEMORY_FUNCS memFuncs = { appMalloc, appFree, appRealloc, appCalloc, NULL  };

/*
 * Init CSSM; returns CSSM_FALSE on error. Reusable.
 */
static CSSM_BOOL cssmInitd = CSSM_FALSE;
CSSM_RETURN cssmStartup()
{
	if(cssmInitd) {
		return CSSM_OK;
	}  
	
	CSSM_RETURN  crtn;
    CSSM_PVC_MODE pvcPolicy = CSSM_PVC_NONE;
	
	crtn = CSSM_Init (&vers, 
					  CSSM_PRIVILEGE_SCOPE_NONE,
					  &testGuid,
					  CSSM_KEY_HIERARCHY_NONE,
					  &pvcPolicy,
					  NULL /* reserved */);
	if(crtn != CSSM_OK) {
		return crtn;
	}
	else {
		cssmInitd = CSSM_TRUE;
		return CSSM_OK;
	}
}
/*
 * Cook up a symmetric encryption context for the specified key,
 * inferring all needed attributes solely from the key algorithm.
 * This is obviously not a one-size-fits all function, but rather
 * the "most common case". If you need to encrypt/decrypt with other
 * padding, mode, etc., do it yourself.
 */
static CSSM_RETURN genCryptHandle(
								  CSSM_CSP_HANDLE cspHandle,
								  const CSSM_KEY	*key,
								  const CSSM_DATA	*ivPtr,
								  CSSM_CC_HANDLE	*ccHandle)
{
	CSSM_ALGORITHMS		keyAlg = key->KeyHeader.AlgorithmId;
	CSSM_ALGORITHMS		encrAlg;
	CSSM_ENCRYPT_MODE	encrMode = CSSM_ALGMODE_NONE;
	CSSM_PADDING 		encrPad = CSSM_PADDING_PKCS1;
	CSSM_RETURN			crtn;
	CSSM_CC_HANDLE		ccHand = 0;
	CSSM_ACCESS_CREDENTIALS	creds;
	CSSM_BOOL			isSymmetric = CSSM_TRUE;
	
	/* 
	 * Infer algorithm - ususally it's the same as in the key itself
	 */
	switch(keyAlg) {
		case CSSM_ALGID_3DES_3KEY:
			encrAlg = CSSM_ALGID_3DES_3KEY_EDE;
			break;
		default:
			encrAlg = keyAlg;
			break;
	}
	
	/* infer mode and padding */
	switch(encrAlg) {
			/* 8-byte block ciphers */
		case CSSM_ALGID_DES:
		case CSSM_ALGID_3DES_3KEY_EDE:
		case CSSM_ALGID_RC5:
		case CSSM_ALGID_RC2:
			encrMode = CSSM_ALGMODE_CBCPadIV8;
			encrPad = CSSM_PADDING_PKCS5;
			break;
			
			/* 16-byte block ciphers */
		case CSSM_ALGID_AES:
			encrMode = CSSM_ALGMODE_CBCPadIV8;
			encrPad = CSSM_PADDING_PKCS7;
			break;
			
			/* stream ciphers */
		case CSSM_ALGID_ASC:
		case CSSM_ALGID_RC4:
			encrMode = CSSM_ALGMODE_NONE;
			encrPad = CSSM_PADDING_NONE;
			break;
			
			/* RSA asymmetric */
		case CSSM_ALGID_RSA:
			/* encrMode not used */
			encrPad = CSSM_PADDING_PKCS1;
			isSymmetric = CSSM_FALSE;
			break;
		default:
			/* don't wing it - abort */
			return CSSMERR_CSP_INTERNAL_ERROR;
	}
	
	memset(&creds, 0, sizeof(CSSM_ACCESS_CREDENTIALS));
	if(isSymmetric) {
		crtn = CSSM_CSP_CreateSymmetricContext(cspHandle,
											   encrAlg,
											   encrMode,
											   NULL,			// access cred
											   key,
											   ivPtr,			// InitVector
											   encrPad,
											   NULL,			// Params
											   &ccHand);
	}
	else {
		crtn = CSSM_CSP_CreateAsymmetricContext(cspHandle,
												encrAlg,
												&creds,			// access
												key,
												encrPad,
												&ccHand);
		
	}
	if(crtn) {
		return crtn;
	}
	*ccHandle = ccHand;
	return CSSM_OK;
}

/*
 * Initialize CDSA and attach to the CSP.
 */
CSSM_RETURN cdsaCspAttach(
						  CSSM_CSP_HANDLE		*cspHandle)
{
	CSSM_CSP_HANDLE cspHand;
	CSSM_RETURN		crtn;
	
	/* initialize CDSA (this is reusable) */
	crtn = cssmStartup();
	if(crtn) {
		return crtn;
	}
	
	/* Load the CSP bundle into this app's memory space */
	crtn = CSSM_ModuleLoad(&gGuidAppleCSP,
						   CSSM_KEY_HIERARCHY_NONE,
						   NULL,			// eventHandler
						   NULL);			// AppNotifyCallbackCtx
	if(crtn) {
		return crtn;
	}
	
	/* obtain a handle which will be used to refer to the CSP */ 
	crtn = CSSM_ModuleAttach (&gGuidAppleCSP,
							  &vers,
							  &memFuncs,			// memFuncs
							  0,					// SubserviceID
							  CSSM_SERVICE_CSP,	
							  0,					// AttachFlags
							  CSSM_KEY_HIERARCHY_NONE,
							  NULL,				// FunctionTable
							  0,					// NumFuncTable
							  NULL,				// reserved
							  &cspHand);
	if(crtn) {
		return crtn;
	}
	*cspHandle = cspHand;
	return CSSM_OK;
}

/*
 * Detach from CSP. To be called when app is finished with this 
 * library.
 */
CSSM_RETURN cdsaCspDetach(
						  CSSM_CSP_HANDLE		cspHandle)
{
	return CSSM_ModuleDetach(cspHandle);
}

#pragma mark ------ Key generation ------

/*
 * Derive a symmetric CSSM_KEY from the specified raw key material.
 */
CSSM_RETURN cdsaDeriveKey(
						  CSSM_CSP_HANDLE		cspHandle,
						  const void 			*rawKey,
						  size_t				rawKeyLen,
						  CSSM_ALGORITHMS		keyAlg,			// e.g., CSSM_ALGID_AES
						  uint32				keySizeInBits,
						  CSSM_KEY_PTR		key)
{
	CSSM_RETURN					crtn;
	CSSM_CC_HANDLE 				ccHand;
	CSSM_DATA					dummyLabel = {8, (uint8 *)"sidecar"};
	CSSM_DATA					saltData = {8, (uint8 *)"gonfishn"};
	CSSM_PKCS5_PBKDF2_PARAMS 	pbeParams;
	CSSM_DATA					pbeData;
	CSSM_ACCESS_CREDENTIALS		creds;
	
	memset(key, 0, sizeof(CSSM_KEY));
	memset(&creds, 0, sizeof(CSSM_ACCESS_CREDENTIALS));
	crtn = CSSM_CSP_CreateDeriveKeyContext(cspHandle,
										   CSSM_ALGID_PKCS5_PBKDF2,
										   keyAlg,
										   keySizeInBits,
										   &creds,
										   NULL,			// BaseKey
										   1000,			// iterationCount, 1000 is the minimum
										   &saltData,
										   NULL,			// seed
										   &ccHand);
	if(crtn) {
		return crtn;
	}
	
	/* this is the caller's raw key bits, typically ASCII (though it
	 * could be anything) */
	pbeParams.Passphrase.Data = (uint8 *)rawKey;
	pbeParams.Passphrase.Length = rawKeyLen;
	/* The only PRF supported by the CSP is HMACSHA1 */
	pbeParams.PseudoRandomFunction = CSSM_PKCS5_PBKDF2_PRF_HMAC_SHA1;
	pbeData.Data = (uint8 *)&pbeParams;
	pbeData.Length = sizeof(pbeParams);
	crtn = CSSM_DeriveKey(ccHand,
						  &pbeData,
						  CSSM_KEYUSE_ANY,
						  CSSM_KEYATTR_RETURN_DATA | CSSM_KEYATTR_EXTRACTABLE,
						  &dummyLabel,
						  NULL,			// cred and acl
						  key);
	CSSM_DeleteContext(ccHand);		// ignore error here
	return crtn;
}

/*
 * Free resources allocated in cdsaDeriveKey().
 */
CSSM_RETURN cdsaFreeKey(
						CSSM_CSP_HANDLE		cspHandle,
						CSSM_KEY_PTR		key)
{
	return CSSM_FreeKey(cspHandle, 
						NULL,			// access cred
						key,	
						CSSM_FALSE);	// don't delete since it wasn't permanent
}

#pragma mark ------ Simple encrypt/decrypt routines ------

/* 
 * Common initialization vector shared by encrypt and decrypt.
 * Some applications may wish to specify a different IV for
 * each encryption op (e.g., disk block number, IP packet number,
 * etc.) but that is outside the scope of this library.
 */
static uint8 iv[16] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };
static const CSSM_DATA ivCommon = {16, iv};

/*
 * Encrypt.
 * cipherText->Data is allocated by the CSP and must be freed (via
 * free()) by caller.
 */
CSSM_RETURN cdsaEncrypt(
						CSSM_CSP_HANDLE		cspHandle,
						const CSSM_KEY		*key,
						const CSSM_DATA		*plainText,
						CSSM_DATA_PTR		cipherText)
{
	CSSM_RETURN 	crtn;
	CSSM_CC_HANDLE	ccHandle;
	CSSM_DATA		remData = {0, NULL};
	CSSM_SIZE			bytesEncrypted;
	
	crtn = genCryptHandle(cspHandle, key, &ivCommon, &ccHandle);
	if(crtn) {
		return crtn;
	}
	cipherText->Length = 0;
	cipherText->Data = NULL;
	crtn = CSSM_EncryptData(ccHandle,
							plainText,
							1,
							cipherText,
							1,
							&bytesEncrypted,
							&remData);
	CSSM_DeleteContext(ccHandle);
	if(crtn) {
		return crtn;
	}
	
	cipherText->Length = bytesEncrypted;
	if(remData.Length != 0) {
		/* append remaining data to cipherText */
		uint32 newLen = cipherText->Length + remData.Length;
		cipherText->Data = (uint8 *)appRealloc(cipherText->Data,
											   newLen,
											   NULL);
		memmove(cipherText->Data + cipherText->Length, 
				remData.Data, remData.Length);
		cipherText->Length = newLen;
		appFree(remData.Data, NULL);
	}
	return CSSM_OK;
}

/*
 * Decrypt.
 * plainText->Data is allocated by the CSP and must be freed (via
 * free()) by caller.
 */
CSSM_RETURN cdsaDecrypt(
						CSSM_CSP_HANDLE		cspHandle,
						const CSSM_KEY		*key,
						const CSSM_DATA		*cipherText,
						CSSM_DATA_PTR		plainText)
{
	CSSM_RETURN 	crtn;
	CSSM_CC_HANDLE	ccHandle;
	CSSM_DATA		remData = {0, NULL};
	CSSM_SIZE			bytesDecrypted;
	
	crtn = genCryptHandle(cspHandle, key, &ivCommon, &ccHandle);
	if(crtn) {
		return crtn;
	}
	plainText->Length = 0;
	plainText->Data = NULL;
	crtn = CSSM_DecryptData(ccHandle,
							cipherText,
							1,
							plainText,
							1,
							&bytesDecrypted,
							&remData);
    int unwired = mlock(plainText->Data, plainText->Length);
    if(unwired)
    {
        fatal(9); 
    }
	CSSM_DeleteContext(ccHandle);
	if(crtn) {
		return crtn;
	}
	
	plainText->Length = bytesDecrypted;
	if(remData.Length != 0) {
		/* append remaining data to plainText */
		uint32 newLen = plainText->Length + remData.Length;
		plainText->Data = (uint8 *)appRealloc(plainText->Data,
											  newLen,
											  NULL);
		memmove(plainText->Data + plainText->Length, 
				remData.Data, remData.Length);
		plainText->Length = newLen;
		appFree(remData.Data, NULL);
	}
	return CSSM_OK;
}
#pragma mark METHODS
+(NSData *)securityProcessedDataForData:(NSData *)inData direction:(BOOL)decrypting
{
    id outData = nil;
	
    if(inData)
    {
        uint32				keySizeInBits=128;
        const char			*kstr;              // UTF8 password
        CSSM_CSP_HANDLE		cspHandle;
        CSSM_RETURN			crtn;
        CSSM_DATA			inStuff ;				// data to decrypt
        CSSM_DATA			outStuff = {0, NULL};// restored data
        CSSM_KEY			cdsaKey;
        CSSM_ALGORITHMS		encrAlg = CSSM_ALGID_AES;
        static const char * abstract[] = {
			"71b27f34e120646f6840e24b5c8dead9",
			"3c1b9d3ffbece4268d41969b44f95d68",
			"85a97e2d6be1d181e5c0266a44c0915f",
			"67157f94e61aa9bc420bbace1ca02624",
			"d429ad2c32da41999dbb5f8761374982",
			"75b7cfebbc8e16a19cc6382d40cce32a",/*g*/
			"b6b03c3ee84af7ce312ad001713954ac",
			"ca2c349a100349d692a20f8363bf8f32",
			0};
		
        
        // lock the key in memory to prevent it from paging out
        int keyLock = mlock(&cdsaKey, sizeof(cdsaKey));
        if(keyLock)
        {
            fprintf(stderr,"DEBUG: failed to lock key (%s);\n",strerror(errno));
        }
        
		
        
        kstr = abstract[4];
        kstr = abstract[5];
		//    kstr+=3;
		
        
        /* attach to CSP */
        crtn = cdsaCspAttach(&cspHandle);
        if(crtn)
        {
            fatal(19);
        }
		
        /*
         * Derive an actual encryption/decryption key from the password 
         * ASCII text. We could use the ASCII text directly as key 
         * material but using the DeriveKey function is much more secure 
         * (besides being an industry-standard way to convert an ASCII 
         * password into binary key material). 
         */
        crtn = cdsaDeriveKey(cspHandle, kstr, strlen(kstr), encrAlg, keySizeInBits, &cdsaKey);
        if(crtn)
        {
            fatal(18);
        }
        
        inStuff.Data = (uint8 *)[inData bytes];
        inStuff.Length = [inData length];
		
        if(decrypting)
        {
            crtn = cdsaDecrypt(cspHandle, &cdsaKey, &inStuff, &outStuff);
        }
        else
        {
            crtn = cdsaEncrypt(cspHandle, &cdsaKey, &inStuff, &outStuff);
        }
        if(crtn)
        {
            showErr(crtn);
            //fatal(17);
        }
        else if(outStuff.Data)
        {
            outData = [NSData dataWithBytesNoCopy:outStuff.Data length:outStuff.Length freeWhenDone:YES];
        }
        // free up the resources 
        cdsaFreeKey(cspHandle, &cdsaKey);
        bzero(&cdsaKey, sizeof(cdsaKey));
        munlock(&cdsaKey, sizeof(cdsaKey));
        cdsaCspDetach(cspHandle);
    }
    
    return outData;
}
+(id)decryptedPropertyListObjectForData:(NSData *)encryptedData
{
    id plistObject = [self securityProcessedDataForData:encryptedData direction:YES];
    if(plistObject)
    {
        plistObject = [(id)CFPropertyListCreateFromXMLData(NULL, (CFDataRef)plistObject, kCFPropertyListMutableContainersAndLeaves, NULL) autorelease];
    }
    return plistObject;
}
+(NSData *)encryptedDataForPropertyListObject:(id)plistObject
{
    NSData * encryptedData = nil;
    if(plistObject)
    {
        encryptedData = [(NSData *)CFPropertyListCreateXMLData(NULL, plistObject) autorelease];
        //now encrypt the data
        encryptedData = [self securityProcessedDataForData:encryptedData direction:NO];
    }
    [[NSMutableDictionary dictionary] setObject:plistObject forKey:plistObject];
    return encryptedData;
}
+ (void)initialize
{
    if(self == [ATMPPR_SecureSettings class])
    {
        g_mutablePersistentDict = nil;
        g_sharedInstance = nil;
        g_didRestoreFromPrevious = NO;
        g_lastError = nil;
    }
}
-(BOOL) didRestoreFromPrevious
{
    return didRestoreFromPrevious_;
}
-(void)dealloc
{
    if (self.needsSaving == YES)
    {
        [self save];
    }
    [filePath_ release];
    [dataStore_ release];
    [lastError_ release];
    [super dealloc];
}
-(id)initWithPath:(NSString *)filePath
{
    if ((self = [super init]))
    {
        didRestoreFromPrevious_ = NO;
        self.needsSaving = NO;
		
        self.privateDataStore = nil;
        self.privateFilePath = filePath;
        NSError * e = nil;	// 2012-02-09
        id plistObject = nil;
        NSData * diskData = [NSData dataWithContentsOfFile:filePath options:0 error:&e];
		plistObject = [[self class] decryptedPropertyListObjectForData:diskData];
        if([diskData length] && (plistObject = [[self	class] decryptedPropertyListObjectForData:diskData]) && ( [plistObject respondsToSelector:@selector(objectForKey:)] && [plistObject respondsToSelector:@selector(setObject: forKey:)] ))
        {
			self.privateDataStore = plistObject;
			didRestoreFromPrevious_ = YES;
        }
        else
        {
            //try to create a file
            NSMutableDictionary * m_d = [NSMutableDictionary dictionaryWithCapacity:5];
            if([m_d writeToFile:filePath atomically:YES] && (0 == (chmod([filePath fileSystemRepresentation], S_IRUSR | S_IWUSR))))
            {
                self.privateDataStore = m_d;
                
            }
            else
            {
                // failed to own the file. Bail and set error
                e = [NSError errorWithDomain:NSPOSIXErrorDomain code:errno userInfo:nil];
                g_lastError = e.retain;
            }
        }
        if (nil == self.privateDataStore)
        {
            //we failed. dealloc and return nil
            g_lastError = e.retain;
            //if(![e.domain isEqualToString:NSCocoaErrorDomain] || e.code != NSFileNoSuchFileError)
            [self release];
            self = nil;
        }
    }
    return self;
}
+(NSError *)lastError
{
    return g_lastError;
}
-(NSError *)lastError
{
    return self.privateLastError;
}
-(NSString *) filePath
{
    return self.privateFilePath;
}
-(NSMutableDictionary *) dataStore
{
    return self.privateDataStore;
}
+ (id)persistentSettings
{
    if( nil == g_sharedInstance )
    {
        g_sharedInstance = [[[self class] alloc] init];
    }
    
    return g_sharedInstance;
}
+(NSString *)internalDataStorePath
{
    NSString * savedPath = @"/var/tmp/";
    savedPath = [savedPath stringByAppendingPathComponent:[NSString stringWithFormat:@"%@-%u",[[NSBundle mainBundle] bundleIdentifier],[[[NSBundle mainBundle] bundleIdentifier] hash]]];
    return savedPath;
}
+ (void)saveSettingsOrError:(NSError **)error
{
    if( nil != g_mutablePersistentDict)
    {
        NSString * savedPath = [self internalDataStorePath];
        NSData * encryptedData = [self encryptedDataForPropertyListObject:g_mutablePersistentDict];
        if([encryptedData writeToFile:savedPath options:0 error:error])
        {
            // set the mode for universal access
            chmod([savedPath fileSystemRepresentation], S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR | S_IWGRP| S_IWOTH);
        }
    }
}
- (BOOL)save
{
    BOOL didSave = NO;
    if (self.privateDataStore)
    {
        if(self.needsSaving)
        {
            NSData * encryptedData = [[self	class] encryptedDataForPropertyListObject:self.privateDataStore];
            NSError * e;
            if([encryptedData writeToFile:self.privateFilePath options:NSDataWritingAtomic error:&e])
            {
                self.needsSaving = NO;
                didSave = YES;
            }
            else
            {
                self.privateLastError = e;
            }
        }
    }
    return didSave;
}
+ (void)saveSettings
{
    [self saveSettingsOrError:NULL];
}
+ (id)internalDataStore
{
    // NEVER CALL THIS METHOD OR DIRECTLY USE the GLOBAL VARIABLE THAT IT RETURNS!!!
    // ALWAYS USE THE ACCESS METHODS TO SET OR GET A VALUE/KEY COMBINATION!!!
    
    // return the shared persistent dictionary as recovered from disk
    if( nil == g_mutablePersistentDict)
    {
        NSString * savedPath = [self internalDataStorePath];
        db("using persistent settings saved path \"%@\"", savedPath);
        NSError * e;
        id plistObject = nil;
        NSData * diskData = [NSData dataWithContentsOfFile:savedPath options:0 error:&e];
        if(diskData)
        {
            plistObject = [self decryptedPropertyListObjectForData:diskData];
        }
        if( [plistObject respondsToSelector:@selector(objectForKey:)] && [plistObject respondsToSelector:@selector(setObject: forKey:)] )
        {
            g_mutablePersistentDict = [plistObject retain];
            g_didRestoreFromPrevious = YES;
        }
        else
        {
            g_mutablePersistentDict = [[NSMutableDictionary alloc] initWithCapacity:5];
        }
    }
    return g_mutablePersistentDict;
}
// these methods call into the persistent data store which is the decrypted NSMutableDictionary
- (id)objectForKey:(NSString *)key error:(NSError **)e
{
    id object = nil;
    id internalDataStore = [[self class] internalDataStore];
    if([internalDataStore respondsToSelector:@selector(objectForKey:)])
    {
        object = [internalDataStore objectForKey:key];
    }
    else if([internalDataStore respondsToSelector:@selector(valueForKey:)])
    {
        object = [internalDataStore valueForKey:key];
    }
    else if( NULL != e )
    {
        * e = [NSError errorWithDomain:NSCocoaErrorDomain code:NSPropertyListReadCorruptError userInfo:[NSDictionary dictionaryWithObjectsAndKeys:NSLocalizedString(@"internal data failure",""), NSLocalizedDescriptionKey,
																										NSLocalizedString(@"internal failure: persistent settings not a dictionary-class object",""), NSLocalizedFailureReasonErrorKey, nil]];
    }
    return object;
}
- (id)objectForKey:(NSString *)key
{
    if (self.privateDataStore)
    {
        return [self.privateDataStore objectForKey:key];
    }
    else
    {
        return [self objectForKey:key error:NULL];
    }
}
+ (void)setObject:(id)value forKey:(NSString *)key error:(NSError **)e
{
    if([value isKindOfClass:[NSDictionary class]] || [value isKindOfClass:[NSString class]] || [value isKindOfClass:[NSData class]] || [value isKindOfClass:[NSDate class]] || [value isKindOfClass:[NSNumber class]] || [value isKindOfClass:[NSArray class]] )
    {
        ;
    }
    else
    {
        db("WARNING: setting a non-property-list object with class \"%@\" will probably not work!", NSStringFromClass([value class]));
    }
    
    id internalDataStore = [self internalDataStore];
    if([internalDataStore respondsToSelector:@selector(setObject: forKey:)])
    {
        [internalDataStore setObject:value forKey:key];
    }
    else if([internalDataStore respondsToSelector:@selector(setValue: forKey:)])
    {
        [internalDataStore setValue:value forKey:key];
    }
    else if( NULL != e )
    {
        * e = [NSError errorWithDomain:@"errordomain" code:0 userInfo:[NSDictionary dictionaryWithObjectsAndKeys:NSLocalizedString(@"internal data failure",""), NSLocalizedDescriptionKey,
																	   NSLocalizedString(@"internal failure: persistent settings not a dictionary-class object",""), NSLocalizedFailureReasonErrorKey,nil]];
    }
    [self saveSettings];
}
+ (void)setObject:(id)value forKey:(NSString *)key
{
    [self setObject:value forKey:key error:NULL];
}
- (void)setObject:(id)value forKey:(NSString *)key
{
    if (self.privateDataStore)
    {
        [self.privateDataStore setObject:value forKey:key];
        self.needsSaving = YES;
    }
    else
    {
        [[self class] setObject:value forKey:key];
    }
}
+ (void)removeObjectForKey:(NSString *)key
{
    id internalDataStore = [self internalDataStore];
    if([internalDataStore respondsToSelector:@selector(removeObjectForKey:)])
    {
        [internalDataStore removeObjectForKey:key];
    }
    [self saveSettings];    
}
- (void)removeObjectForKey:(NSString *)key
{
    if (self.privateDataStore)
    {
        [self.privateDataStore removeObjectForKey:key];
        self.needsSaving = YES;
    }
    else
    {
        [[self class] removeObjectForKey:key];
    }
}
+(BOOL)validatedFromPreviousSession
{
    [self internalDataStore];
    return g_didRestoreFromPrevious;
}
-(BOOL)validatedFromPreviousSession
{
	return [[self class] validatedFromPreviousSession];
}
@end
